#include "StdAfx.h"
#include "NMMemDC.h"


NMMemDC::NMMemDC()
{
    m_hOldBitmap=NULL;
    m_pvBits = NULL;
    m_bAlphaChannel=false;
}

NMMemDC::~NMMemDC()
{
    DeleteDC();
}

void NMMemDC::CreateCompatibleDC(CDC* pDC,UINT nWidth,UINT nHeight,bool bAlphaChannel)
{
    CDC::CreateCompatibleDC(pDC);
    //m_bmpMem.DeleteObject();

    m_nBmpWidth=nWidth;
    m_nBmpHeight=nHeight;

    m_bAlphaChannel=bAlphaChannel;
    if(!bAlphaChannel)
    {
        m_bmpMem.CreateCompatibleBitmap(pDC,nWidth,nHeight);
    }
    else
    {
	    BITMAPINFOHEADER bmih = { 0 };   
	    //int nBytesPerLine = ((nWidth * 32 + 31) & (~31)) >> 3;
	    // Populate BITMAPINFO header   
	    bmih.biSize = sizeof(BITMAPINFOHEADER);   
	    bmih.biWidth = nWidth;   
	    bmih.biHeight = nHeight;   
	    bmih.biPlanes = 1;   
	    bmih.biBitCount = 32;   
	    bmih.biCompression = BI_RGB;   
	    bmih.biClrUsed = 0;   
	    //bmih.biSizeImage = nBytesPerLine * nHeight;   
        bmih.biSizeImage = 0;   

	    m_pvBits = NULL;   
	    HBITMAP hbmpMem = ::CreateDIBSection(NULL, (PBITMAPINFO)&bmih, DIB_RGB_COLORS, &m_pvBits, NULL, 0);
	    
        ASSERT(hbmpMem != NULL);
        if(hbmpMem==NULL)
            return;

	    memset( m_pvBits, 0, nWidth * 4 * nHeight);

        m_bmpMem.Attach(hbmpMem);
    }
    m_hOldBitmap=SelectObject(&m_bmpMem);
}

void NMMemDC::CreateCompatibleDC(CDC* pDC,LPRECT lpRect,bool bAlphaChannel)
{
    CreateCompatibleDC(pDC,lpRect->right-lpRect->left,lpRect->bottom-lpRect->top,bAlphaChannel);
}

BOOL NMMemDC::DeleteDC()
{
    if(m_bmpMem.GetSafeHandle())
    {
        if(m_hDC)
        {
            SelectObject(m_hOldBitmap);
        }
        m_bmpMem.DeleteObject();
    }
    return CDC::DeleteDC();
}

HBITMAP NMMemDC::GetBitmap()
{
    return (HBITMAP)m_bmpMem.GetSafeHandle();
}

HBITMAP NMMemDC::DetachBitmap()
{
    SelectObject(m_hOldBitmap);
    return (HBITMAP)m_bmpMem.Detach();
}

LPBYTE NMMemDC::GetBmpBits()
{
    if(m_bAlphaChannel)
        return (LPBYTE)m_pvBits;
    else
        return NULL;
}

bool NMMemDC::AlphaTransparentBlt(int xDest, int yDest,
        CImage& image,int xSrc, int ySrc, int nSrcWidth, int nSrcHeight)
{
    if(image.IsNull())
        return false;

    if(!m_bAlphaChannel)
        return false;
    
    LPBYTE lpDstBits=(LPBYTE)m_pvBits;
    LPBYTE lpSrcBits=(LPBYTE)image.GetBits();

    int nCountPerDCLine=m_nBmpWidth*4;
    int nCountPerImageLine=0;

    int nPitchCount=image.GetPitch();
    if(nPitchCount<0)
    {   //image bootom-up image
        //dst top-down image̴.
        lpSrcBits = lpSrcBits+((image.GetHeight()-1) * nPitchCount);
        nCountPerImageLine=-nPitchCount;
    }
    else
    {
        nCountPerImageLine=nPitchCount;
    }

    if( (xDest>=m_nBmpWidth)||(yDest>=m_nBmpHeight) )
        return true;

    LPBYTE dr, dg, db, da, sr, sg, sb, sa;
    for(int y=0, nDestYPos=yDest ; y<nSrcHeight && nDestYPos<m_nBmpHeight ; y++,nDestYPos++)
    {
        register LPBYTE lpDst=lpDstBits+(nDestYPos*nCountPerDCLine)+xDest*4;
        register LPBYTE lpSrc=lpSrcBits+((y+ySrc)*nCountPerImageLine)+xSrc*4;
        for(int x=0, nDestXPos=xDest ; x<nSrcWidth && nDestXPos<m_nBmpWidth; x++,nDestXPos++)
        {
            dr = lpDst++;
            dg = lpDst++;
            db = lpDst++;
            da = lpDst++;
            sr = lpSrc++;
            sg = lpSrc++;
            sb = lpSrc++;
            sa = lpSrc++;

            if(*sa > 0)
            {
#if 1
                *dr=*sr;
                *dg=*sg;
                *db=*sb;
                *da=*sa;
                //*da=255;
#else
                *dr=(BYTE)(((UINT)(*sr) * (*sa))/255);
                *dg=(BYTE)(((UINT)(*sg) * (*sa))/255);
                *db=(BYTE)(((UINT)(*sb) * (*sa))/255);
#endif
            }
        }
    }

    return true;
}

bool NMMemDC::CopyAlphaImage(int xDest, int yDest,
    CImage& image,int xSrc, int ySrc, int nSrcWidth, int nSrcHeight,COLORREF crBkgnd)
{
    if(image.IsNull())
        return false;

    if(!m_bAlphaChannel)
        return false;
    
    LPBYTE lpDstBits=(LPBYTE)m_pvBits;
    LPBYTE lpSrcBits=(LPBYTE)image.GetBits();

    int nCountPerDCLine=m_nBmpWidth*4;
    int nCountPerImageLine=0;

    int nPitchCount=image.GetPitch();
#if 1
    if(nPitchCount<0)
    {   //image bootom-up image
        //dst top-down image̴.
        nCountPerImageLine=-nPitchCount;
        lpSrcBits = lpSrcBits+((image.GetHeight()-1) * nPitchCount);
    }
    else
    {
        nCountPerImageLine=nPitchCount;
    }
#else
    nCountPerImageLine=nPitchCount;
#endif

    if( (xDest>=m_nBmpWidth)||(yDest>=m_nBmpHeight) )
        return true;

    LPBYTE dr, dg, db, da, sr, sg, sb, sa;
    for(int y=0, nDestYPos=yDest ; y<nSrcHeight && nDestYPos<m_nBmpHeight ; y++,nDestYPos++)
    {
        register LPBYTE lpDst=lpDstBits+(nDestYPos*nCountPerDCLine)+xDest*4;
        register LPBYTE lpSrc=lpSrcBits+((y+ySrc)*nCountPerImageLine)+xSrc*4;
        for(int x=0, nDestXPos=xDest ; x<nSrcWidth && nDestXPos<m_nBmpWidth; x++,nDestXPos++)
        {
            dr = lpDst++;
            dg = lpDst++;
            db = lpDst++;
            da = lpDst++;
            sr = lpSrc++;
            sg = lpSrc++;
            sb = lpSrc++;
            sa = lpSrc++;
#if 1
            if(*sa > 0)
            {
#pragma warning(disable: 4244)
                double fOpacity=*sa/255.0;
                *dr=*sr * fOpacity + (1.0-fOpacity) * GetRValue(crBkgnd);
                *dg=*sg * fOpacity + (1.0-fOpacity) * GetGValue(crBkgnd);
                *db=*sb * fOpacity + (1.0-fOpacity) * GetBValue(crBkgnd);
                *da=*sa * fOpacity + (1.0-fOpacity) * *da;
                //*dr=*sr + (1-*sa) * GetRValue(crBkgnd);
                //*dg=*sg + (1-*sa) * GetGValue(crBkgnd);
                //*db=*sb + (1-*sa) * GetBValue(crBkgnd);
                //*da=*sa + (1-*sa) * *da;
#pragma warning(default: 4244)
            }
#else
            *dr=*sr + (1-*sa)*(*dr);
            *dg=*sg + (1-*sa)*(*dg);
            *db=*sb + (1-*sa)*(*db);
            *da=*sa + (1-*sa)*(*da);
#endif
        }
    }

    return true;
}

bool NMMemDC::FillImageRegion(int xDest, int yDest,
    CImage& image,int xSrc, int ySrc, int nSrcWidth, int nSrcHeight,COLORREF crBkgnd)
{
    if(image.IsNull())
        return false;

    if(!m_bAlphaChannel)
        return false;
    
    LPBYTE lpDstBits=(LPBYTE)m_pvBits;
    LPBYTE lpSrcBits=(LPBYTE)image.GetBits();

    int nCountPerDCLine=m_nBmpWidth*4;
    int nCountPerImageLine=0;

    int nPitchCount=image.GetPitch();
    if(nPitchCount<0)
    {   //image bootom-up image
        //dst top-down image̴.
        lpSrcBits = lpSrcBits+((image.GetHeight()-1) * nPitchCount);
        nCountPerImageLine=-nPitchCount;
    }
    else
    {
        nCountPerImageLine=nPitchCount;
    }

    if( (xDest>=m_nBmpWidth)||(yDest>=m_nBmpHeight) )
        return true;

    register LPBYTE dr, dg, db, da, sr, sg, sb, sa;
    for(int y=0, nDestYPos=yDest ; y<nSrcHeight && nDestYPos<m_nBmpHeight ; y++,nDestYPos++)
    {
        register LPBYTE lpDst=lpDstBits+(nDestYPos*nCountPerDCLine)+xDest*4;
        register LPBYTE lpSrc=lpSrcBits+((y+ySrc)*nCountPerImageLine)+xSrc*4;

        for(int x=0, nDestXPos=xDest ; x<nSrcWidth && nDestXPos<m_nBmpWidth; x++,nDestXPos++)
        {
            dr = lpDst++;
            dg = lpDst++;
            db = lpDst++;
            da = lpDst++;
            sr = lpSrc++;
            sg = lpSrc++;
            sb = lpSrc++;
            sa = lpSrc++;

            *dr=GetRValue(crBkgnd);
            *dg=GetGValue(crBkgnd);
            *db=GetBValue(crBkgnd);
            *da=0;
        }
    }

    return true;
}

bool NMMemDC::FillImageRegion(NMMemDC& dcSrc,COLORREF crBkgnd)
{
    LPBYTE lpDstBits=(LPBYTE)m_pvBits;
    LPBYTE lpSrcBits=(LPBYTE)dcSrc.GetBmpBits();
    if(!lpDstBits || !lpSrcBits)
        return false;

    LPBYTE dr, dg, db, da, sr, sg, sb, sa;
    for(int y=0 ; y<m_nBmpHeight; y++)
    {
        for(int x=0; x<m_nBmpWidth; x++)
        {
            dr = lpDstBits++;
            dg = lpDstBits++;
            db = lpDstBits++;
            da = lpDstBits++;
            sr = lpSrcBits++;
            sg = lpSrcBits++;
            sb = lpSrcBits++;
            sa = lpSrcBits++;

            if(*sa>0)
            {
                *dr=GetRValue(crBkgnd);
                *dg=GetGValue(crBkgnd);
                *db=GetBValue(crBkgnd);
                *da=0;
            }
        }
    }

    return true;
}

void NMMemDC::ClearAlphaValue()
{
    if(!m_bAlphaChannel)
        return;

    LPBYTE lpBits=(LPBYTE)m_pvBits;
    LPBYTE r,g,b,a;
    for(int y=0;y<m_nBmpHeight;y++)
    {
        for(int x=0;x<m_nBmpWidth;x++)
        {
            r=lpBits++;
            g=lpBits++;
            b=lpBits++;
            a=lpBits++;

            *a=0;
        }
    }
}

bool NMMemDC::TransparentBltToAlphaLayer(NMMemDC& dcSrc,COLORREF crTrans,int nAlphaVlaue)
{
    BITMAP bmInfoSrc,bmInfoDest;
    GetCurrentBitmap()->GetBitmap(&bmInfoDest);
    dcSrc.GetCurrentBitmap()->GetBitmap(&bmInfoSrc);
    
    if(bmInfoDest.bmBitsPixel!=32 || bmInfoSrc.bmBitsPixel!=32)
    {
        ASSERT(0);
        return false;
    }

    LPBYTE lpDstBits=(LPBYTE)m_pvBits;
    LPBYTE lpSrcBits=(LPBYTE)dcSrc.GetBmpBits();
    if(!lpDstBits || !lpSrcBits)
        return false;

    LPBYTE dr, dg, db, da, sr, sg, sb, sa;
    for(int y=0 ; y<m_nBmpHeight; y++)
    {
        for(int x=0; x<m_nBmpWidth; x++)
        {
            dr = lpDstBits++;
            dg = lpDstBits++;
            db = lpDstBits++;
            da = lpDstBits++;
            sr = lpSrcBits++;
            sg = lpSrcBits++;
            sb = lpSrcBits++;
            sa = lpSrcBits++;

            if(crTrans != RGB(*sr,*sg,*sb))
            {
                *dr=*sr;
                *dg=*sg;
                *db=*sb;
                if(nAlphaVlaue>=0)
                    *da=nAlphaVlaue;
            }
        }
    }

    return true;    
}